@@ -14,7 +14,7 @@
 //
 // We mean it.
 
-#include <private/qtqmlcompilerexports_p.h>
+#include <qtqmlcompilerexports.h>
 
 #include <QtCore/qstring.h>
 #include <QtCore/qstringlist.h>
@@ -59,7 +59,8 @@ class QQmlJSMetaEnum
     QString m_typeName;
     QSharedPointer<const QQmlJSScope> m_type;
     bool m_isFlag = false;
-    bool m_scoped = true;
+    bool m_isScoped = false;
+    bool m_isQml = false;
 
 public:
     QQmlJSMetaEnum() = default;
@@ -76,8 +77,11 @@ public:
     bool isFlag() const { return m_isFlag; }
     void setIsFlag(bool isFlag) { m_isFlag = isFlag; }
 
-    bool isScoped() const { return m_scoped; }
-    void setScoped(bool v) { m_scoped = v; }
+    bool isScoped() const { return m_isScoped; }
+    void setIsScoped(bool v) { m_isScoped = v; }
+
+    bool isQml() const { return m_isQml; }
+    void setIsQml(bool v) { m_isQml = v; }
 
     void addKey(const QString &key) { m_keys.append(key); }
     QStringList keys() const { return m_keys; }
@@ -102,7 +106,8 @@ public:
                 && a.m_name == b.m_name
                 && a.m_alias == b.m_alias
                 && a.m_isFlag == b.m_isFlag
-                && a.m_type == b.m_type;
+                && a.m_type == b.m_type
+                && a.m_isScoped == b.m_isScoped;
     }
 
     friend bool operator!=(const QQmlJSMetaEnum &a, const QQmlJSMetaEnum &b)
@@ -112,7 +117,8 @@ public:
 
     friend size_t qHash(const QQmlJSMetaEnum &e, size_t seed = 0)
     {
-        return qHashMulti(seed, e.m_keys, e.m_values, e.m_name, e.m_alias, e.m_isFlag, e.m_type);
+        return qHashMulti(
+                seed, e.m_keys, e.m_values, e.m_name, e.m_alias, e.m_isFlag, e.m_type, e.m_isScoped);
     }
 };
 
@@ -130,10 +136,13 @@ public:
         Const,
     };
 
-    QQmlJSMetaParameter(const QString &name, const QString &typeName,
+    QQmlJSMetaParameter(QString name = QString(), QString typeName = QString(),
                         Constness typeQualifier = NonConst,
                         QWeakPointer<const QQmlJSScope> type = {})
-        : m_name(name), m_typeName(typeName), m_type(type), m_typeQualifier(typeQualifier)
+        : m_name(std::move(name)),
+          m_typeName(std::move(typeName)),
+          m_type(type),
+          m_typeQualifier(typeQualifier)
     {
     }
 
@@ -153,7 +162,7 @@ public:
     friend bool operator==(const QQmlJSMetaParameter &a, const QQmlJSMetaParameter &b)
     {
         return a.m_name == b.m_name && a.m_typeName == b.m_typeName
-                && a.m_type.toStrongRef().data() == b.m_type.toStrongRef().data()
+                && a.m_type.owner_equal(b.m_type)
                 && a.m_typeQualifier == b.m_typeQualifier;
     }
 
@@ -164,7 +173,7 @@ public:
 
     friend size_t qHash(const QQmlJSMetaParameter &e, size_t seed = 0)
     {
-        return qHashMulti(seed, e.m_name, e.m_typeName, e.m_type.toStrongRef().data(),
+        return qHashMulti(seed, e.m_name, e.m_typeName, e.m_type.owner_hash(),
                           e.m_typeQualifier);
     }
 
@@ -177,6 +186,8 @@ private:
     bool m_isList = false;
 };
 
+using QQmlJSMetaReturnType = QQmlJSMetaParameter;
+
 class QQmlJSMetaMethod
 {
 public:
@@ -203,22 +214,29 @@ public:
     QQmlJSMetaMethod() = default;
     explicit QQmlJSMetaMethod(QString name, QString returnType = QString())
         : m_name(std::move(name)),
-          m_returnTypeName(std::move(returnType)),
+          m_returnType(QString(), std::move(returnType)),
           m_methodType(MethodType::Method)
     {}
 
     QString methodName() const { return m_name; }
     void setMethodName(const QString &name) { m_name = name; }
 
-    QString returnTypeName() const { return m_returnTypeName; }
-    QSharedPointer<const QQmlJSScope> returnType() const { return m_returnType.toStrongRef(); }
-    void setReturnTypeName(const QString &type) { m_returnTypeName = type; }
-    void setReturnType(const QSharedPointer<const QQmlJSScope> &type)
-    {
-        m_returnType = type;
-    }
+    QQmlJS::SourceLocation sourceLocation() const { return m_sourceLocation; }
+    void setSourceLocation(QQmlJS::SourceLocation location) { m_sourceLocation = location; }
+
+    QQmlJSMetaReturnType returnValue() const { return m_returnType; }
+    void setReturnValue(const QQmlJSMetaReturnType returnValue) { m_returnType = returnValue; }
+    QString returnTypeName() const { return m_returnType.typeName(); }
+    void setReturnTypeName(const QString &typeName) { m_returnType.setTypeName(typeName); }
+    QSharedPointer<const QQmlJSScope> returnType() const { return m_returnType.type(); }
+    void setReturnType(QWeakPointer<const QQmlJSScope> type) { m_returnType.setType(type); }
 
     QList<QQmlJSMetaParameter> parameters() const { return m_parameters; }
+    QPair<QList<QQmlJSMetaParameter>::iterator, QList<QQmlJSMetaParameter>::iterator>
+    mutableParametersRange()
+    {
+        return { m_parameters.begin(), m_parameters.end() };
+    }
 
     QStringList parameterNames() const
     {
@@ -290,11 +308,10 @@ public:
 
     friend bool operator==(const QQmlJSMetaMethod &a, const QQmlJSMetaMethod &b)
     {
-        return a.m_name == b.m_name && a.m_returnTypeName == b.m_returnTypeName
-                && a.m_returnType == b.m_returnType && a.m_parameters == b.m_parameters
-                && a.m_annotations == b.m_annotations && a.m_methodType == b.m_methodType
-                && a.m_methodAccess == b.m_methodAccess && a.m_revision == b.m_revision
-                && a.m_isConstructor == b.m_isConstructor;
+        return a.m_name == b.m_name && a.m_returnType == b.m_returnType
+                && a.m_parameters == b.m_parameters && a.m_annotations == b.m_annotations
+                && a.m_methodType == b.m_methodType && a.m_methodAccess == b.m_methodAccess
+                && a.m_revision == b.m_revision && a.m_isConstructor == b.m_isConstructor;
     }
 
     friend bool operator!=(const QQmlJSMetaMethod &a, const QQmlJSMetaMethod &b)
@@ -307,8 +324,7 @@ public:
         QtPrivate::QHashCombine combine;
 
         seed = combine(seed, method.m_name);
-        seed = combine(seed, method.m_returnTypeName);
-        seed = combine(seed, method.m_returnType.toStrongRef().data());
+        seed = combine(seed, method.m_returnType);
         seed = combine(seed, method.m_annotations);
         seed = combine(seed, method.m_methodType);
         seed = combine(seed, method.m_methodAccess);
@@ -324,9 +340,10 @@ public:
 
 private:
     QString m_name;
-    QString m_returnTypeName;
-    QWeakPointer<const QQmlJSScope> m_returnType;
 
+    QQmlJS::SourceLocation m_sourceLocation;
+
+    QQmlJSMetaReturnType m_returnType;
     QList<QQmlJSMetaParameter> m_parameters;
     QList<QQmlJSAnnotation> m_annotations;
 
@@ -426,7 +443,7 @@ public:
     {
         return a.m_index == b.m_index && a.m_propertyName == b.m_propertyName
                 && a.m_typeName == b.m_typeName && a.m_bindable == b.m_bindable
-                && a.m_type == b.m_type && a.m_isList == b.m_isList
+                && a.m_type.owner_equal(b.m_type) && a.m_isList == b.m_isList
                 && a.m_isWritable == b.m_isWritable && a.m_isPointer == b.m_isPointer
                 && a.m_aliasExpr == b.m_aliasExpr && a.m_revision == b.m_revision
                 && a.m_isFinal == b.m_isFinal;
@@ -455,7 +472,7 @@ public:
     create a new binding, you know all the details of it already, so you should
     just set all the data at once.
 */
-class Q_QMLCOMPILER_PRIVATE_EXPORT QQmlJSMetaPropertyBinding
+class Q_QMLCOMPILER_EXPORT QQmlJSMetaPropertyBinding
 {
     using BindingType = QQmlSA::BindingType;
     using ScriptBindingKind = QQmlSA::ScriptBindingKind;
@@ -524,7 +541,7 @@ class Q_QMLCOMPILER_PRIVATE_EXPORT QQmlJSMetaPropertyBinding
             ScriptBindingValueType valueType = ScriptBindingValueType::ScriptValue_Unknown;
         };
         struct Object {
-            friend bool operator==(Object a, Object b) { return a.value == b.value && a.typeName == b.typeName; }
+            friend bool operator==(Object a, Object b) { return a.value.owner_equal(b.value) && a.typeName == b.typeName; }
             friend bool operator!=(Object a, Object b) { return !(a == b); }
             QString typeName;
             QWeakPointer<const QQmlJSScope> value;
@@ -532,7 +549,7 @@ class Q_QMLCOMPILER_PRIVATE_EXPORT QQmlJSMetaPropertyBinding
         struct Interceptor {
             friend bool operator==(Interceptor a, Interceptor b)
             {
-                return a.value == b.value && a.typeName == b.typeName;
+                return a.value.owner_equal(b.value) && a.typeName == b.typeName;
             }
             friend bool operator!=(Interceptor a, Interceptor b) { return !(a == b); }
             QString typeName;
@@ -541,7 +558,7 @@ class Q_QMLCOMPILER_PRIVATE_EXPORT QQmlJSMetaPropertyBinding
         struct ValueSource {
             friend bool operator==(ValueSource a, ValueSource b)
             {
-                return a.value == b.value && a.typeName == b.typeName;
+                return a.value.owner_equal(b.value) && a.typeName == b.typeName;
             }
             friend bool operator!=(ValueSource a, ValueSource b) { return !(a == b); }
             QString typeName;
@@ -570,7 +587,7 @@ class Q_QMLCOMPILER_PRIVATE_EXPORT QQmlJSMetaPropertyBinding
             */
             friend bool operator==(AttachedProperty a, AttachedProperty b)
             {
-                return a.value == b.value;
+                return a.value.owner_equal(b.value);
             }
             friend bool operator!=(AttachedProperty a, AttachedProperty b) { return !(a == b); }
             QWeakPointer<const QQmlJSScope> value;
@@ -592,7 +609,7 @@ class Q_QMLCOMPILER_PRIVATE_EXPORT QQmlJSMetaPropertyBinding
                ### TODO: Obtaining the effective binding result requires some resolving function
             */
             QWeakPointer<const QQmlJSScope> groupScope;
-            friend bool operator==(GroupProperty a, GroupProperty b) { return a.groupScope == b.groupScope; }
+            friend bool operator==(GroupProperty a, GroupProperty b) { return a.groupScope.owner_equal(b.groupScope); }
             friend bool operator!=(GroupProperty a, GroupProperty b) { return !(a == b); }
         };
         using type = std::variant<Invalid, BoolLiteral, NumberLiteral, StringLiteral,
@@ -725,8 +742,6 @@ public:
         m_bindingContent = Content::ValueSource { typeName, type };
     }
 
-    QString literalTypeName() const;
-
     // ### TODO: here and below: Introduce an allowConversion parameter, if yes, enable conversions e.g. bool -> number?
     bool boolValue() const;
 
@@ -860,7 +875,7 @@ public:
     }
 };
 
-struct Q_QMLCOMPILER_PRIVATE_EXPORT QQmlJSMetaSignalHandler
+struct Q_QMLCOMPILER_EXPORT QQmlJSMetaSignalHandler
 {
     QStringList signalParameters;
     bool isMultiline;
